# Basic imports
import numpy as np
import pandas as pd
from   typing import Any
from   numpy.typing import ArrayLike
from   cachai._core.chord import ChordDiagram
from   cachai import utilities as util
# Matplotlib imports
from   matplotlib import pyplot as plt
from   matplotlib.axes import Axes

def chord(
    corr_matrix      : np.ndarray | pd.DataFrame,
    names            : ArrayLike = None,
    colors           : ArrayLike = None,
    *,
    ax               : Axes = None,
    radius           : float = 1,
    position         : ArrayLike = (0,0),
    optimize         : bool = True,
    filter           : bool = True,
    bezier_n         : int = 30,
    show_diag        : bool = False,
    threshold        : float = 0.1,
    node_linewidth   : float = 10,
    node_gap         : float = 0.1,
    node_labelpad    : float = 0.2,
    blend            : bool = True,
    blend_resolution : int = 200,
    chord_linewidth  : float = 1,
    chord_alpha      : float = 0.7,
    off_alpha        : float = 0.1,
    positive_hatch   : str = None,
    negative_hatch   : str = '---',
    fontsize         : int = 15,
    font             : dict | str = None,
    min_dist         : float = np.deg2rad(15),
    scale            : str = 'linear',
    max_rho          : float = 0.4,
    max_rho_radius   : float = 0.7,
    show_axis        : bool = False,
    legend           : bool = False,
    positive_label   : str = None,
    negative_label   : str = None,
    rasterized       : bool = False,
    **kwargs         : Any,
    ) -> ChordDiagram:
    """
    Create and return a ChordDiagram visualization.
    
    Parameters:
    -----------
    corr_matrix : numpy.ndarray or pandas.DataFrame
        Correlation matrix for the chord diagram
    names / n : list
        Names for each node (default: 'Ni' for the i-th node)
    colors / c : list
        Custom colors for nodes (default: seaborn hls palette)
    ax : matplotlib.axes.Axes
        Axes to plot on (default: current pyplot axis)
    radius / r : float
        Radius of the diagram (default: 1.0)
    position / p : tuple
        Position of the center of the diagram (default: (0,0))
    optimize : bool
        Whether to optimize node order (default: True)
    filter : bool
        Whether to remove nodes with no correlation (default: True)
    bezier_n : int
        Bezier curve resolution (default: 30)
    show_diag : bool
        Show self-connections (default: False)
    threshold / th : float
        Minimum correlation threshold to display (default: 0.1)
    node_linewidth / nlw : float
        Line width for nodes (default: 10)
    node_gap / ngap : float
        Gap between nodes (0-1) (default: 0.1)
    node_labelpad / npad : float
        Label position adjustment (default: 0.2)
    blend : bool
        Whether to blend chord colors (default: True)
    blend_resolution : int
        Color blend resolution (default: 200)
    chord_linewidth / clw : float
        Line width for chords (default: 1)
    chord_alpha / calpha : float
        Alpha of the facecolor for chords (default: 0.7)
    off_alpha : float
        Alpha for non-highlighted chords (default: 0.1)
    positive_hatch : str
        Hatch for positive correlated chords (default: None)
    negative_hatch : str
        Hatch for negative correlated chords (default: '---')
    fontsize : int
        Label font size (default: 15)
    font : dict or str
        Label font parameters (default: None)
    min_dist : float
        Minimum angle distance from which apply radius rule (default: 15 [degrees])
    scale : str
        Scale use to set chord's thickness, wheter "linear" or "log" (default: "linear")
    max_rho : float
        Maximum chord's thickness (default: 0.4) 
    max_rho_radius : float
        Maximum normalized radius of the chords relative to center (default: 0.7)
    show_axis : bool
        Whether to show the axis (default: False)
    legend : bool
        Adds default positive and negative labels in the legend (default: False)
    positive_label : str
        Adds positive label in the legend (default: None)
    negative_label : str
        Adds negative label in the legend (default: None)
    rasterized : bool
        Whether to force rasterized (bitmap) drawing for vector graphics output (default: False)
    
    Returns:
    --------
    ChordDiagram
        An instance of the ChordDiagram class
    """
    # Process parameters
    params = {
        'corr_matrix'      : corr_matrix,
        'names'            : names,
        'colors'           : colors,
        'ax'               : ax if ax is not None else plt.gca(),
        'radius'           : radius,
        'position'         : position,
        'optimize'         : optimize,
        'filter'           : filter,
        'bezier_n'         : bezier_n,
        'show_diag'        : show_diag,
        'threshold'        : threshold,
        'node_linewidth'   : node_linewidth,
        'node_gap'         : node_gap,
        'node_labelpad'    : node_labelpad,
        'blend'            : blend,
        'blend_resolution' : blend_resolution,
        'chord_linewidth'  : chord_linewidth,
        'chord_alpha'      : chord_alpha,
        'off_alpha'        : off_alpha,
        'positive_hatch'   : positive_hatch,
        'negative_hatch'   : negative_hatch,
        'fontsize'         : fontsize,
        'font'             : font,
        'min_dist'         : min_dist,
        'scale'            : scale,
        'max_rho'          : max_rho,
        'max_rho_radius'   : max_rho_radius,
        'show_axis'        : show_axis,
        'legend'           : legend,
        'positive_label'   : positive_label,
        'negative_label'   : negative_label,
        'rasterized'       : rasterized,
    }
    
    # Alternative kwargs aliases
    aliases = {'names'           : 'n',
               'colors'          : 'c',
               'radius'          : 'r',
               'position'        : 'p',
               'threshold'       : 'th',
               'node_linewidth'  : 'nlw',
               'node_gap'        : 'ngap',
               'node_labelpad'   : 'npad',
               'chord_linewidth' : 'clw',
               'chord_alpha'     : 'calpha'
              }
    
    for key in aliases:
        if aliases[key] in kwargs: params[key] = kwargs.pop(aliases[key])

    # Check for wrong kwargs
    util.validate_kwargs(params,kwargs,aliases)

    return ChordDiagram(**params)

    


# Basic imports
import numpy as np
import pandas as pd
import seaborn as sns
from   cachai import utilities as util
from   cachai import gadgets
# Matplotlib imports
from   matplotlib import pyplot as plt
from   matplotlib.patches import Arc, Circle, PathPatch
from   matplotlib.path import Path
import matplotlib.colors as mtpl_colors
from   matplotlib.text import Text


class ChordDiagram():
    def __init__(self, corr_matrix, **kwargs):
        """
        Initialize a ChordDiagram visualization.
        """
        # Correlation matrix error handling
        self.corr_matrix = corr_matrix
        self._validate_corr_matrix()
        
        # Initialize additional parameters
        self.__dict__.update(kwargs)
        if isinstance(self.corr_matrix, pd.DataFrame):
            if self.names is None: self.names = self.corr_matrix.columns.tolist()
            self.corr_matrix = self.corr_matrix.to_numpy()
        if self.names is None: self.names = [f'N{i+1}' for i in range(len(self.corr_matrix))]
        if self.colors is None: self.colors = sns.hls_palette(len(self.corr_matrix))
        self.nodes = dict()
        self.order = [i for i in range(len(self.corr_matrix))]
        self.global_indexes = []
        if self.font is None: self.font = {'size':self.fontsize}
        
        # Initialize collection lists
        self.node_patches        = []
        self.node_labels         = []
        self.node_labels_params  = []
        self.chord_patches       = [[] for i in range(len(self.corr_matrix))]
        self.chord_blends        = [[] for i in range(len(self.corr_matrix))]
        self.bezier_curves       = [[] for i in range(len(self.corr_matrix))]
        self.__ports_refs        = []
        self.__highlighted_ports = []
        
        # Generate the diagram
        self.__generate_diagram()
    
    
    # Util methods
    def _validate_corr_matrix(self):
        """Validate that a correlation matrix meets the required specifications:
            - Input is a numpy.ndarray or pandas.DataFrame
            - Matrix is 2-dimensional
            - Matrix is not empty
            - All values are int or float
            - Matrix is symmetric
        """
        temp_corr_matrix = self.corr_matrix
        if not isinstance(temp_corr_matrix, (np.ndarray, pd.DataFrame)):
            raise TypeError('Your correlation matrix must be a numpy.ndarray or pandas.DataFrame')
        # -- This block of code should not be here, but its necessary for the next validations --
        if isinstance(temp_corr_matrix, pd.DataFrame):
            temp_corr_matrix = self.corr_matrix.to_numpy()
        # ---------------------------------------------------------------------------------------
        if temp_corr_matrix.ndim != 2:
            raise ValueError('Your correlation matrix must be a 2-dimensional array')
        if temp_corr_matrix.shape[0] != temp_corr_matrix.shape[1]:
            raise ValueError('Your correlation matrix must be a square matrix.')
        if len(temp_corr_matrix) == 0:
            raise ValueError('Your correlation matrix cannot be empty')
        if not np.issubdtype(temp_corr_matrix.dtype, np.floating):
            raise TypeError('Your correlation matrix must contain float values')
        if not np.allclose(temp_corr_matrix, temp_corr_matrix.T):
            raise ValueError('Your correlation matrix must be symmetric')
        
    def _optimize_nodes(self):
        """Optimize node order using Prim's algorithm."""
        n_nodes = self.corr_matrix.shape[0]
        # We convert the correlations to distances
        # The strongest the correlation, the shorter the distance
        distance_matrix = 1 - np.abs(self.corr_matrix)
        # In order to ignore the diagonal, we fill it with infinity values
        np.fill_diagonal(distance_matrix, np.inf)
        
        # Prim's algorithm
        visited = set()
        order  = []

        start_node = np.argmin(np.sum(distance_matrix, axis=0))
        visited.add(start_node)
        order.append(start_node)

        while len(visited) < n_nodes:
            # Closest non-visited node to any visited node
            min_dist = np.inf
            next_node = -1
            for node in visited:
                for neighbor in range(n_nodes):
                    if (neighbor not in visited) and (distance_matrix[node,neighbor] < min_dist):
                        min_dist = distance_matrix[node, neighbor]
                        next_node = neighbor
            if next_node == -1:
                break  # Just in case somehow we have disconnected nodes
            visited.add(next_node)
            order.append(next_node)

        # Apply new order
        self.order = order
        self.__order_nodes()
    
    def _radius_rule(self, dist):
        """Rule to set the radius of a single chord"""
        if dist <= self.min_dist:
            return self.max_rho_radius
        else:
            return self.max_rho_radius * (1 - (dist - self.min_dist) / (np.pi - self.min_dist))
    
    def _scale_rho(self, rho):
        """Scale rho (link thickness)"""
        if self.scale == 'linear':
            rho_lin = np.abs(rho) * self.max_rho
            return np.clip(rho_lin, 0, 1) # Clip to avoid numerical issues
        elif self.scale == 'log':
            rho_log = (1 - np.log10(10 - 9*np.abs(rho))) * self.max_rho
            return np.clip(rho_log, 0, 1) # Clip to avoid numerical issues
        else:
            raise ValueError(f'Unknown scale type {self.scale}')
    
    # Main generation methods
    def __generate_diagram(self):
        """Generate the complete chord diagram"""
        
        for i,color in enumerate(self.colors):
            if isinstance(color,str): self.colors[i] = mtpl_colors.to_rgb(color)
            
        if self.filter == True: self.__filter_nodes()
        
        if len(self.corr_matrix) == 0:
            raise ValueError(f'No nodes remaining after threshold filtering: '
                f'all correlations were below the threshold = {self.threshold}.')
        else:
            if self.optimize == True: self._optimize_nodes()
            self.__generate_nodes()
            self.__generate_chords()

            # Add patches to axes
            for node_patch, node_label in zip(self.node_patches, self.node_labels_params):
                self.ax.add_patch(node_patch)
                label = gadgets.PolarText(self.position,
                                          node_label['r'],
                                          node_label['theta'],
                                          text=node_label['label'],
                                          pad=self.node_labelpad,
                                          rotation=node_label['rot'],
                                          ha='center', va='center',
                                          clip_on=True,
                                          rasterized=self.rasterized)
                label.set_font(self.font)
                self.ax.add_artist(label)
                self.node_labels.append(label)
            flat_chord_patches = [p for plist in self.chord_patches for p in plist]
            flat_bezier_curves = [c for clist in self.bezier_curves for c in clist]
            for k,(chord_patch,bezier_curve) in enumerate(zip(flat_chord_patches,flat_bezier_curves)):
                self.ax.add_patch(chord_patch)
                if self.blend:
                    self.__add_chord_blend(chord_patch,bezier_curve,self.global_indexes[k])
            
            self.__adjust_ax()
            self.__generate_legend()
            self.__generate_port_refs()
                
    # Components generation methods
    def __generate_nodes(self):
        """Generate nodes"""
        # Initial variables
        # Minus 1 from each diagonal of A to A
        relevance      = np.sum(np.abs(self.corr_matrix),axis=1) - 1
        relevance_norm = relevance / np.sum(relevance)
        start_angles   = np.cumsum([0] + list(2*np.pi*relevance_norm[:-1]))
        gap_angle      = (2*np.pi/len(self.corr_matrix))*self.node_gap

        # Base patch
        self.ax.add_patch(Circle(self.position,self.radius,
                                 lw=0,
                                 zorder=2,
                                 fc='w',
                                 ec='none',
                                 rasterized=self.rasterized))

        lw = 2*self.node_linewidth

        for node in range(len(self.corr_matrix)):
            node_data = dict()
            theta_i   = start_angles[node]
            theta_f   = theta_i + 2*np.pi*relevance_norm[node]
            # -- Gap correction -----
            theta_i   = theta_i + np.min([gap_angle,theta_f])
            # -----------------------
            theta_m   = (theta_i + theta_f)/2
            theta_arc = util.angdist(theta_i,theta_f)
            node_data['theta_i']   = theta_i
            node_data['theta_f']   = theta_f
            node_data['theta_m']   = theta_m
            node_data['theta_arc'] = theta_arc

            rhos       = dict() # Correlations
            ports      = dict() # Ports of the node
            states     = dict() # Ports states (1 or -1)
            corr       = np.insert(self.corr_matrix[node],node,1)
            real_corr  = corr.copy()
            # Control of the allowed ports using the correlation factor
            # 1 = Allowed
            # -1 = Forbidden
            node_ports_state = [1 for p in range(len(corr))]
            for p,r in enumerate(corr):
                if np.abs(r) < self.threshold:
                    node_ports_state[p] = -1
                    real_corr[p]        = 0
            if not self.show_diag:
                node_ports_state[node]   = -1
                real_corr[node]          = 0
                node_ports_state[node+1] = -1
                real_corr[node+1]        = 0
            if np.sum(real_corr) == 0: corr_norm = real_corr
            else: corr_norm = np.abs(real_corr)/np.sum(np.abs(real_corr))

            for j,(rho,port_size,port_state) in enumerate(zip(corr,corr_norm,node_ports_state)):
                if   j == node     : port_id = node
                elif j == (node+1) : port_id = f'{node}*'
                else:
                    if   j < node: port_id = j
                    elif j > node: port_id = j-1
                port_i = theta_i + theta_arc*np.sum(corr_norm[:j])
                port_f = port_i + theta_arc*port_size
                if port_state < 0:
                    port_i = 0
                    port_f = 0
                rhos[port_id]   = rho
                ports[port_id]  = {'i':port_i,'f':port_f}
                states[port_id] = port_state

            node_data['rhos']        = rhos
            node_data['ports']       = ports
            node_data['ports_state'] = states
            self.nodes[node]         = node_data

        # Node
        for n in self.nodes:
            node = self.nodes[n]

            # Patch
            self.node_patches.append(
                Arc(self.position,
                width=2*self.radius, 
                height=2*self.radius,
                theta1=np.rad2deg(node['theta_i']), 
                theta2=np.rad2deg(node['theta_f']),
                lw=lw,zorder=1, 
                rasterized=self.rasterized,
                color=self.colors[n])
            )

            # Label
            params = dict()
            params['label'] = self.names[n]
            params['r']     = self.radius
            params['theta'] = node['theta_m']
            params['x']     = params['r'] * np.cos(params['theta']) + self.position[0]
            params['y']     = params['r'] * np.sin(params['theta']) + self.position[1]
            params['rot']   = np.rad2deg(node['theta_m'] - np.sign(params['y']-self.position[1])*np.pi/2)%360
            self.node_labels_params.append(params)
        
    def __generate_chords(self):
        """Generate chords"""
        for n in self.nodes:
            node = self.nodes[n]
            chord_color = self.colors[n]
            chord_edge  = util.mod_color(self.colors[n],light=0.5)
            if self.blend:
                chord_color = 'none'
                chord_edge = '#3D3D3D'

            # Links
            if self.show_diag:
                points,codes,curve = self.__compute_bezier_curves(
                                         (node['ports'][n]['i'],node['ports'][n]['f']),
                                         (node['ports'][f'{n}*']['i'],node['ports'][f'{n}*']['f']),
                                         self._scale_rho(1)
                                     )

                self.chord_patches[n].append(
                    PathPatch(Path(points, codes),
                              facecolor=chord_color,
                              edgecolor=chord_edge,
                              alpha=self.chord_alpha,
                              hatch=self.positive_hatch,
                              lw=self.chord_linewidth,
                              rasterized=self.rasterized,
                              zorder=4)
                )
                curve['c1'] = self.colors[n]
                curve['c2'] = self.colors[n]
                self.bezier_curves[n].append(curve)
                self.global_indexes.append(n)

            for m in self.nodes:
                if m > n and node['ports_state'][m] > 0:
                    try:
                        target   = self.nodes[m]
                        this_rho = node['rhos'][m]
                        vis_rho  = self._scale_rho(this_rho)
                        hatch    = self.positive_hatch
                        if this_rho < 0: hatch = self.negative_hatch

                        points,codes,curve = self.__compute_bezier_curves(
                                         (node['ports'][m]['i'],node['ports'][m]['f']),
                                         (target['ports'][n]['i'],target['ports'][n]['f']),
                                         vis_rho
                                     )

                        self.chord_patches[n].append(
                            PathPatch(Path(points, codes),
                                      facecolor=chord_color,
                                      edgecolor=chord_edge,
                                      alpha=self.chord_alpha,
                                      hatch=hatch,
                                      lw=self.chord_linewidth,
                                      rasterized=self.rasterized,
                                      zorder=4)
                        )
                        curve['c1'] = self.colors[n]
                        curve['c2'] = self.colors[m]
                        self.bezier_curves[n].append(curve)
                        self.global_indexes.append(n)
                        
                    except Exception as e:
                        print(util.textco(rf'ChordError: Problem creating chord from {self.names[n]} to {self.names[m]}.',
                                          c='red'))
                        print(util.textco(f'            details: {e}',
                                          c='red'))
    
    def __generate_legend(self):
        """Add dummie labels to show in the legend"""
        if self.legend is True:
            if self.positive_label is None: self.positive_label = 'Positive\ncorrelation'
            if self.negative_label is None: self.negative_label = 'Negative\ncorrelation'
        # Dummies
        if self.positive_label is not None:
            dummy = self.ax.scatter(*self.position,marker='s',s=200,
                                    c='lightgray',ec='k',hatch=self.positive_hatch,
                                    label=self.positive_label,zorder=0,rasterized=True)
            #dummy.set_visible(False)
        if self.negative_label is not None:
            dummy = self.ax.scatter(*self.position,marker='s',s=200,
                                    c='lightgray',ec='k',hatch=self.negative_hatch,
                                    label=self.negative_label,zorder=0,rasterized=True)
            #dummy.set_visible(False)
    
    def __generate_port_refs(self):
        for n in self.nodes:
            self.__ports_refs.append(self.__get_node_ports_references(n))

    # Helper methods
    def __filter_nodes(self):
        """Remove nodes with no correlation (0 chords)"""
        mask = np.all((np.abs(self.corr_matrix) < self.threshold)\
                      | (np.eye(self.corr_matrix.shape[0], dtype=bool)), axis=1)
        indexes = np.where(~mask)[0]
        
        self.corr_matrix = self.corr_matrix[indexes][:, indexes]
        self.names       = [self.names[i] for i in indexes]
        self.colors      = [self.colors[i] for i in indexes]
        
    def __order_nodes(self):
        """Order nodes (matrix), names and colors"""
        self.corr_matrix = self.corr_matrix[np.ix_(self.order, self.order)]
        self.names       = [self.names[i] for i in self.order]
        self.colors      = [self.colors[i] for i in self.order]
    
    def __compute_bezier_curves(self,alpha,beta,rho):
        """Compute bezier curves to modelate a chord"""
        # Polar
        alpha_i, alpha_f = alpha
        alpha_m = np.mean([alpha_f,alpha_i])
        alphas = util.angspace(alpha_i,alpha_f)
        if len(alphas) == 0: alphas = np.array([alpha_i,alpha_f]) # Case: angdist too short

        beta_i, beta_f = beta
        beta_m = np.mean([beta_f,beta_i])
        betas = util.angspace(beta_i,beta_f)
        if len(betas) == 0: betas = np.array([beta_i,beta_f]) # Case: angdist too short

        dist = util.angdist(alpha_m, beta_m)
        r_rho = self._radius_rule(dist) * self.radius
        dist_inex = np.min([util.angdist(alpha_i, beta_f), util.angdist(alpha_f, beta_i)])

        # Convex case
        if util.angdist(alpha_i, beta_f) < util.angdist(alpha_f, beta_i):
            theta_rho = beta_f + dist_inex / 2
            r_AB = r_rho
            r_BA = r_rho + rho * self.radius
        # Concave case
        elif util.angdist(alpha_i, beta_f) >= util.angdist(alpha_f, beta_i):
            theta_rho = alpha_f + dist_inex / 2
            r_AB = r_rho + rho * self.radius
            r_BA = r_rho

        # Cartesian
        points_A = np.column_stack([np.cos(alphas) * self.radius, 
                                   np.sin(alphas) * self.radius])
        points_B = np.column_stack([np.cos(betas) * self.radius, 
                                   np.sin(betas) * self.radius])

        # A to B
        point_AB = [np.array([r_AB * np.cos(theta_rho), 
                             r_AB * np.sin(theta_rho)])]
        control_AB = [2 * point_AB[0] - (points_A[-1] + points_B[0]) / 2]

        # B to A
        point_BA = [np.array([r_BA * np.cos(theta_rho), 
                             r_BA * np.sin(theta_rho)])]
        control_BA = [2 * point_BA[0] - (points_A[0] + points_B[-1]) / 2]

        # Bezier curve in the middle
        mid_bezier = dict()
        r_mid = (r_AB + r_BA) / 2
        point_mid = [np.array([r_mid * np.cos(theta_rho), 
                              r_mid * np.sin(theta_rho)])]
        control_mid = [2 * point_mid[0] - (points_A[-1] + points_B[0]) / 2]
        mid_bezier['P0'] = points_A[-1] + self.position
        mid_bezier['P1'] = control_mid[0] + self.position
        mid_bezier['P2'] = points_B[0] + self.position

        # Points
        points = np.vstack((points_A, control_AB, points_B, control_BA, points_A[0])) \
                 + self.position

        # Codes
        codes = [Path.MOVETO] + \
                [Path.LINETO] * (len(points_A) - 1) + \
                [Path.CURVE3] * 2 + \
                [Path.LINETO] * (len(points_B) - 1) + \
                [Path.CURVE3] * 2

        return points,codes,mid_bezier
    
    def __add_chord_blend(self,patch,curve,n):
        """Add color mapped patches using the initial and final colors"""
        # Pach vertices
        vertices = patch.get_path().vertices
        xmin, ymin = np.min(vertices, axis=0)
        xmax, ymax = np.max(vertices, axis=0)

        # Bézier curve
        P0 = curve['P0']
        P1 = curve['P1']
        P2 = curve['P2']
        bezier = util.get_bezier_curve(P0,P1,P2,n=self.bezier_n)
        bezier_equidistant = util.equidistant(bezier)

        # Color map
        norm        = mtpl_colors.Normalize(vmin=-1, vmax=1)
        c1          = curve['c1'] # Color 1
        c2          = curve['c2'] # Color 2
        chord_cmap  = sns.blend_palette([c1,c1,c2,c2],as_cmap=True)
        cmap_matrix = util.map_from_curve(bezier_equidistant,xlims=(xmin,xmax),ylims=(ymin,ymax),
                                     resolution=self.blend_resolution)
        self.chord_blends[n].append(
            util.colormapped_patch(
                patch,
                cmap_matrix,
                ax=self.ax,
                colormap=chord_cmap,
                zorder=2,
                alpha=self.chord_alpha,
                rasterized=self.rasterized)
        )
    
    def __adjust_ax(self):
        """Adjust scale, limits, and visibility of the axis"""
        adjust_x = self.ax.get_autoscalex_on()
        adjust_y = self.ax.get_autoscaley_on()
        if adjust_x:
            self.ax.set_xlim(self.position[0] - self.radius*1.5,self.position[0] + self.radius*1.5)
        if adjust_y:
            self.ax.set_ylim(self.position[1] - self.radius*1.5,self.position[1] + self.radius*1.5)
        if adjust_x and adjust_y: self.ax.set_aspect('equal')
        if self.show_axis == False: self.ax.axis('off')
    
    def __get_node_ports(self,n):
        """Return the occupied ports of the n-th node"""
        this_items = list(self.nodes[n]['ports_state'].items())
        this_items.pop(n+1)
        return [p for p, s in this_items if s > 0 and p > n]

    def __get_node_ports_references(self,n):
        """
        Return the reference of the chords of the n-th node as (n,c), where:

        - n: index of the n-th node in the resulting diagram
        - c: index of the c-th chord in the n-th node

        Always anti-clockwise. When show_diag=True, the self-referencing chord is (n,0).
        """
        node       = self.nodes[n]
        this_ports = self.__get_node_ports(n)
        refs       = []
        for port in node['ports_state'].keys():
            if '*' not in str(port):
                if node['ports_state'][port] > 0 and port > n:
                    refs.append((n,this_ports.index(port)))
                elif node['ports_state'][port] > 0 and port < n:
                    target_ports = self.__get_node_ports(port)
                    refs.append((port,target_ports.index(n)))
        if self.show_diag:
            diag_ref_position = None
            refs_modified = []
            for i, (x, y) in enumerate(refs):
                modified = (x, y + 1)
                refs_modified.append(modified)
                if diag_ref_position is None and x == n: diag_ref_position = i   
            if diag_ref_position is None: diag_ref_position = len(refs_modified)
            refs_modified.insert(diag_ref_position, (n, 0))
            refs = refs_modified
        return refs

    def __update_highlights(self):
        for n in self.nodes:
            for c in range(len(self.chord_patches[n])):
                if (n,c) not in self.__highlighted_ports:
                    self.chord_patches[n][c].set_alpha(self.off_alpha)
                    if self.blend == True: self.chord_blends[n][c].set_alpha(self.off_alpha)

    # Customization methods
    def highlight_node(self,node:int,chords:list=None,alpha:float=None):
        """
        Targeted highlighting of the n-th node
        You can also select the chords that you want to highlight using chords = [0,1,...]
        """
        if node >= len(self.nodes):
            raise IndexError('Node is out of range. '
                f'This Chord Diagram has only {len(self.nodes)} nodes.')
        if node < 0 :
            raise ValueError('The input node must be positive or zero.')

        if chords is None: chords = [c for c in range(len(self.__ports_refs[node]))]
        for chord in chords:
            self.highlight_chord(node,chord,alpha)

    def highlight_chord(self,node:int,chord:int,alpha:float=None):
        """Targeted highlighting of the c-th chord in the n-th node"""
        if node >= len(self.nodes):
            raise IndexError('Node is out of range. '
                f'This Chord Diagram has only {len(self.nodes)} nodes.')
        if node < 0 or chord < 0:
            raise ValueError('The input node and chord must be positive or zero.')

        if alpha is None: alpha = self.chord_alpha
        if alpha <= self.off_alpha: alpha = 0.8

        self.__update_highlights()

        try:
            n,c = self.__ports_refs[node][chord]
            self.chord_patches[n][c].set_alpha(alpha)
            if self.blend == True: self.chord_blends[n][c].set_alpha(alpha)
            if (n,c) not in self.__highlighted_ports: self.__highlighted_ports.append((n,c))
        except IndexError:
            raise IndexError(f'Chord {chord} is out of range. '
                    f'Node {node} has only {len(self.__ports_refs[node])} chords.')

    def set_chord_alpha(self,alpha):
        for n in self.nodes:
            for cp in self.chord_patches[n]: cp.set_alpha(alpha)
            if self.blend == True:
                for cb in self.chord_blends[n]: cb.set_alpha(alpha)

    # Special methods 
    def __str__(self):
        string = ''
        for n in self.nodes:
            string += f'node {n} "{self.names[n]}"\n' + '-'*50
            for key in self.nodes[n]:
                if key == 'ports':
                    string += f'\n{key:<10} :'
                    for p in self.nodes[n][key]:
                        string += f'\n\t\t{p:<10} : {self.nodes[n][key][p]}'
                else:
                    string += f'\n{key:<10} : {self.nodes[n][key]}'
            string += '\n\n\n'
        return string


# Imports
import os
import numpy as np
import colorsys
# Matplotlib imports
from   matplotlib import pyplot as plt
# Scipy imports
from   scipy.spatial.distance import cdist
from   scipy.interpolate import interp1d


def kwargs_as_string(params,aliases):
    """Format arguments into string separated by ,"""
    keys = list(params.keys())
    for i,key in enumerate(params):
        if key in aliases.keys(): keys[i] = f'{key} / {aliases[key]}'
    return ', '.join(keys)

def validate_kwargs(params,kwargs,aliases={}):
    """Check for wrong kwargs"""
    for key in kwargs:
        if key not in params:
            raise KeyError(
            f'Invalid argument "{key}". Allowed arguments are: {kwargs_as_string(params,aliases)}.'
            )

def save_func(fig_name='img',path='images',img_dpi=300,pdf_dpi=200,pdf=True):
    """Save plot as png/pdf"""
    if not os.path.exists(path): os.makedirs(path)
    plt.savefig(os.path.join(path,f'{fig_name}.png'),bbox_inches='tight',pad_inches=0.3,dpi=img_dpi)
    if pdf:
        if not os.path.exists(os.path.join(path,'pdf')): os.makedirs(os.path.join(path,'pdf'))
        plt.savefig(os.path.join(path,'pdf',f'{fig_name}.pdf'),bbox_inches='tight',pad_inches=0.3,dpi=pdf_dpi)

# f-string pre-defined colors
fstr_colors = {'white':255,'black':232,'light_gray':245,'dark_gray':237,'gold':220,
               'red':196,'blue':21,'green':118,'magenta':165,'mint':87,'orange':202}

def color_palette():
    """Print the pyhton color palette (256) for strings"""
    for i in range(16):
        for j in range(16):
            print(textco('■',c=i+j*16) + f' {str(i+j*16):<3.3}  ',end='')
        print()

def textco(text,c='white'):
    """Add color to prints"""
    if isinstance(c,int):
        return f'\033[38;5;{c}m{text}\033[0m'
    else:
        try:
            temp = fstr_colors[c]
        except:
            c = 'white'
        return f'\033[38;5;{fstr_colors[c]}m{text}\033[0m'

def rgb_to_hsl(rgb):
    """Convert color from rgb to hsl"""
    r,g,b = rgb
    h,l,s = colorsys.rgb_to_hls(r,g,b)
    return (h,s,l)

def hsl_to_rgb(hsl):
    """Convert color from hsl to rgb"""
    h,s,l = hsl
    r,g,b = colorsys.hls_to_rgb(h, l, s)
    return (r,g,b)

def saturate_color(color,factor=1):
    """Change saturation of RGB color"""
    h,s,l = rgb_to_hsl(color)
    s_new = max(0, min(1,s*factor))
    r,g,b = hsl_to_rgb((h,s_new,l))
    return (r,g,b)

def alpha_color(color, alpha=1, bg=(1,1,1)):
    """Change alpha of RGB color"""
    r,g,b = color
    bg_r,bg_g,bg_b = bg
    factor = max(0, min(1, alpha))
    r_result = (r*factor) + (bg_r*(1-factor))
    g_result = (g*factor) + (bg_g*(1-factor))
    b_result = (b*factor) + (bg_b*(1-factor))
    return (r_result,g_result,b_result)

def lighter_color(color,factor=0.1):
    """Lighter RGB color"""
    r,g,b = color
    factor = max(0, factor)
    r = min(1, r + (1-r)*factor)
    g = min(1, g + (1-g)*factor)
    b = min(1, b + (1-b)*factor)
    return (r,g,b)

def darker_color(color,factor=0.1):
    """Darker RGB color"""
    r,g,b = color
    factor = 1 - (max(0,factor))
    r = max(0, r*factor)
    g = max(0, g*factor)
    b = max(0, b*factor)
    return (r,g,b)

def mod_color(color,light=1,sat=1,alpha=1,alpha_bg=(1,1,1)):
    """Modify RGB color"""
    new_color = color
    # Light
    if light > 1:
        new_color = lighter_color(new_color,factor=light-1)
    elif light < 1:
        new_color = darker_color(new_color,factor=1-light)
    # Saturation
    new_color = saturate_color(new_color,factor=sat)
    # Alpgha
    new_color = alpha_color(new_color,alpha=alpha,bg=alpha_bg)
    return new_color


def angspace(alpha,beta,n=200):
    """Generate a linear space of angles"""
    theta = abs(beta-alpha)
    ndots = int(theta*n/(2*np.pi))
    if ndots == 1: ndots = 2
    return np.linspace(alpha,beta,ndots)
    
def angdist(alpha,beta):
    """Compute the minimum angular distance"""
    diff = np.abs(alpha - beta) % (2 * np.pi)
    return np.min([diff, 2 * np.pi - diff])

def quadratic_bezier(t,P0:list,P1:list,P2:list):
    """Evaluate quadratic bezier curve in t"""
    if (P0 is None) or (P1 is None) or (P2 is None): return None
    P0 = np.array(P0); P1 = np.array(P1); P2 = np.array(P2)
    return (1-t)**2 * P0 + 2*(1-t)*t * P1 + t**2 * P2

def get_bezier_curve(P0:list,P1:list,P2:list,n=20):
    """Create quadratic bezier curve points [x,y]"""
    if (P0 is None) or (P1 is None) or (P2 is None): return None
    t_values = np.linspace(0, 1, n)
    return np.array([quadratic_bezier(t, P0, P1, P2) for t in t_values])

def equidistant(points):
    """Make any array of points equidistant"""
    # Cumulative distances between consecutive points
    diffs = np.diff(points, axis=0)
    distances = np.linalg.norm(diffs, axis=1)
    cumulative_length = np.insert(np.cumsum(distances), 0, 0)

    # Interpolation
    total_length = cumulative_length[-1]
    new_distances = np.linspace(0, total_length, len(points))

    interp_x = interp1d(cumulative_length, points[:, 0], kind='linear')
    interp_y = interp1d(cumulative_length, points[:, 1], kind='linear')
    
    # Equidistant curve
    new_x = interp_x(new_distances)
    new_y = interp_y(new_distances)
    
    return np.column_stack((new_x, new_y))

def map_from_curve(curve=None,xlims=(-1,1),ylims=(-1,1),resolution=200):
    """Map points using the distance to curve"""
    if curve is None: return None
    
    # Colors values from curve
    colors = np.linspace(-1,1,len(curve))
    
    # Mesh
    x = np.linspace(*xlims, resolution)
    y = np.linspace(*ylims, resolution)
    grid_x, grid_y = np.meshgrid(x, y, indexing='xy')
    grid_points = np.column_stack((grid_x.ravel(), grid_y.ravel()))

    # Obtain the nearest point in the curve and use that color
    distances             = cdist(grid_points, curve)
    nearest_point_indexes = np.argmin(distances, axis=1)
    map_flat              = colors[nearest_point_indexes]
    map_matrix            = map_flat.reshape(resolution, resolution)
    
    return map_matrix

def colormapped_patch(patch,map_matrix,ax=None,colormap='coolwarm',
                      zorder=5,alpha=0.5,rasterized=False):
    """Plot a color coded patch using a map matrix from map_from_curve"""
    if ax is None: ax = plt.gca()
    
    vertices = patch.get_path().vertices
    xmin, ymin = np.min(vertices, axis=0)
    xmax, ymax = np.max(vertices, axis=0)
    
    img = ax.imshow(map_matrix, 
                    cmap=colormap, 
                    extent=(xmin, xmax, ymin, ymax), 
                    origin='lower',
                    aspect='auto',
                    clip_path=patch,
                    clip_on=True,
                    zorder=zorder,
                    alpha=alpha,
                    rasterized=rasterized,
                    vmin=-1, vmax=1)
    return img


import sys
import matplotlib.pyplot as plt
from   pathlib import Path
from   importlib.util import find_spec

def run_tests(*test_args,show_details=True):
    if find_spec('pytest') is None:
        print('\nError: To run the tests you need to have pytest installed.', file=sys.stderr)
        print('You can install it with:', file=sys.stderr)
        print('\n    pip install pytest\n', file=sys.stderr)
        print('Or install cachai with the testing dependency:', file=sys.stderr)
        print('\n    pip install cachai[testing]\n', file=sys.stderr)
        return None

    import pytest
    """
    Run the tests from CACHAI.
    
    Parameters:
    -----------
    *test_args : str
        Names of the tests to run (e.g. 'charts', 'utilities').
        If not specified, all tests are run.
    show_details : bool
        Whether to add -v to the pytest command line (default: True)
    """
    test_dir = Path(__file__).parent
    pytest_args = ['-p', 'no:asdf_schema_tester']
    if show_details:
        pytest_args.append('-v')

    if test_args:
        # Validation
        missing = []
        valid_tests = []
        for arg in test_args:
            test_file = test_dir / f'test_{arg}.py'
            if test_file.exists():
                valid_tests.append(str(test_file))
            else:
                missing.append(arg)
        
        if missing:
            raise ValueError(f'Tests not found: {", ".join(missing)}. '
                           f'The available tests are: {get_available_tests()}')
        
        pytest_args.extend(valid_tests)
    else:
        pytest_args.append(str(test_dir))

    plt.close('all')
    return pytest.main(pytest_args)

def get_available_tests():
    """Returns the list of available tests"""
    test_dir = Path(__file__).parent
    return [f.stem[5:] for f in test_dir.glob('test_*.py')]


# Basic imports
import numpy as np
# Matplotlib imports
from   matplotlib.text import Text

class PolarText(Text):
    def __init__(self, center, radius, angle, text='', pad=0.0, **kwargs):
        """
        Initialize a Text instance using polar coordinates.
        
        Parameters:
        -----------
        center : tuple
            Center of the polar system
        radius : float
            Radius coordinate
        theta : float
            Angle coordinate in rad
        text : str
            Text to diplay (default: '')
        pad : float
            Label position adjustment (default: 0.0)
        
        Keyword Arguments:
        -----------------
        From the parent class Text
        """
        self.center  = np.array(center)
        self._radius = radius
        self._angle  = angle
        self._pad    = pad
        
        x, y = self._polar_to_cartesian(radius*(1 + pad), angle)
        
        super().__init__(x, y, text, **kwargs)

    def _polar_to_cartesian(self, radius, angle):
        """Tranform polar (radius, angle) to catesian (x, y)."""
        dx = radius * np.cos(angle)
        dy = radius * np.sin(angle)
        return self.center + np.array([dx, dy])
    
    def set_pad(self, pad):
        """Update pad"""
        self._pad = pad
        x, y = self._polar_to_cartesian(self._radius*(1 + self._pad), self._angle)
        self.set_position((x, y))
    
    def set_radius(self, radius):
        """Update radius"""
        self._radius = radius
        x, y = self._polar_to_cartesian(self._radius*(1 + self._pad), self._angle)
        self.set_position((x, y))

    def set_angle(self, angle_deg):
        """Update angle in deg"""
        self._angle = np.deg2rad(angle_deg)
        x, y = self._polar_to_cartesian(self._radius*(1 + self._pad), self._angle)
        self.set_position((x, y))

    def set_center(self, center):
        """Update the center coordinates"""
        self.center = np.array(center)
        x, y = self._polar_to_cartesian(self._radius*(1 + self._pad), self._angle)
        self.set_position((x, y))

    def set_polar_position(self, radius=None, angle_deg=None, center=None):
        """Update radius, angle and/or center"""
        if radius is not None:
            self._radius = radius
        if angle_deg is not None:
            self._angle = np.deg2rad(angle_deg)
        if center is not None:
            self.center = np.array(center)
        x, y = self._polar_to_cartesian(self._radius*(1 + self._pad), self._angle)
        self.set_position((x, y))